<?php

namespace xenMade\STWM\Util;

class Watermark
{
    private $_attachment = null;
    private $inputImage;
    public $outputImage;
    private $_attachRepo;
    private $_options;
    private $font = '/styles/default/xenMade/stwm/fonts/NotoSans.ttf';
    private $extensions = array('png', 'jpe', 'jpeg', 'jpg');
    public $status = false;
    protected $_internalPath = null;
    public $isNew = false;

    public function __construct($attachment)
    {
        @ini_set('gd.jpeg_ignore_warning', 1);

        $this->_attachment = $attachment;

        $this->_options = \XF::options();
        $this->_attachRepo = $this->getAttachmentRepo();
        $this->_internalPath = \XF\Util\File::canonicalizePath(\XF::app()->config('internalDataPath'));

        if(empty($this->_options['stwm_watermark_type']['font_path']))
        {
            $this->font = \XF::getRootDirectory() . $this->font;
        }
        else
        {
            $this->font = \XF::getRootDirectory() . $this->_options['stwm_watermark_type']['font_path'];
        }

        if(!$this->_options['stwm_watermark_type']['enabled_image'] && !$this->_options['stwm_watermark_type']['enabled_text'])
            return;

        $metaData = \XF::fs()->getMetadata($attachment->Data->getAbstractedDataPath());

        if(empty($metaData['path']))
            return;

        $this->inputImage = $this->_internalPath . DIRECTORY_SEPARATOR . $metaData['path'];
        $this->outputImage = $this->getWatermarkDataFilePath($attachment);

        // Check Type
        $extension = \XF\Util\File::getFileExtension($attachment['filename']);
        if(!in_array($extension, $this->extensions))
        {
            return;
        }

        // Source Exists?
        if(!file_exists($this->inputImage) || !is_readable($this->inputImage))
            return;

        // Out Jünger als ein Monat?
        if((file_exists($this->outputImage) || is_readable($this->outputImage)) && !\XF::$debugMode)
        {
            if(filemtime($this->outputImage) > time() - (30 * 86400) )
                $this->status = true;
        }

        // Abbruch
        if($this->status)
            return;

        $directory = dirname($this->outputImage);

        if (!file_exists($directory))
        {
            if(!\XF\Util\File::createDirectory($directory, true))
                return;
        }

        try
        {
            $this->generateWatermarkImage();
        }
        catch (\Exception $e)
        {
            if(\XF::$debugMode)
                \XF::app()->logException($e);
        };
    }

    public function generateWatermarkImage()
    {
        if($this->inputImage)
        {
            $imageManager = \XF::app()->imageManager();

            // Fürs Wasserzeichen nur die GD Lib verwenden!!!
            $imageManager->setDefaultDriver('gd');
            $baseImage = $imageManager->imageFromFile($this->inputImage);

            if(!$baseImage instanceof \XF\Image\Gd)
                return;

            if($baseImage->getType() != 2 && $baseImage->getType() != 3)
                return;

            if($baseImage->getWidth() <= 0 || $baseImage->getHeight() <= 0)
                return;

            // Generate WaterText
            if(
                $this->_options['stwm_watermark_type']['enabled_text'] &&

                ($baseImage->getWidth() >= $this->_options['stwm_watermark_type']['text_min_size'] ||
                $baseImage->getHeight() >= $this->_options['stwm_watermark_type']['text_min_size'])
            )
            {
                if(file_exists($this->font) && is_readable($this->font))
                {
                    // Bereite Text vor
                    $text = 'Copyright {year} by {forum}';
                    if(trim($this->_options['stwm_watermark_type']['text_text']))
                    {
                        $text = $this->_options['stwm_watermark_type']['text_text'];
                    }

                    $text = str_replace('{forum}', $this->_options->boardTitle, str_replace('{year}', date("Y"), $text));

                    // Username?
                    if(strpos($text, '{username}') !== false)
                    {
                        $username = $this->_attachRepo->getUserByAttachmentDataId($this->_attachment['data_id']);
                        $text = str_replace('{username}', $username, $text);
                    }

                    // get Text Size and calculate new
                    $fontSize = $this->_options['stwm_watermark_type']['text_size'];

                    $textBox = $this->getTextBox($fontSize, $this->_options['stwm_watermark_type']['text_angle'], $this->font, $text, $baseImage->getWidth(), $baseImage->getHeight());

                    // TMP Image
                    $tmpImage = imagecreatetruecolor($baseImage->getWidth(), $baseImage->getHeight());

                    if($baseImage->getType() === IMAGETYPE_PNG)
                    {
                        if (function_exists('imagealphablending'))
                        {
                            imagealphablending($tmpImage, false);
                            imagesavealpha($tmpImage, true);
                        }
                    }

                    // Copy original to tmp
                    imagecopy($tmpImage, $baseImage->getImage(), 0, 0, 0, 0, $baseImage->getWidth(), $baseImage->getHeight());

                    // Font
                    $textColor = $this->colorToRGB($this->_options['stwm_watermark_type']['text_color']);
                    $textColor = imagecolorallocatealpha($tmpImage, $textColor[0], $textColor[1], $textColor[2], $this->_options['stwm_watermark_type']['text_opacity']);
                    imagecolortransparent($tmpImage, $textColor);

                    if($baseImage->getType() === IMAGETYPE_PNG)
                    {
                        if (function_exists('imagealphablending'))
                        {
                            imagealphablending($tmpImage, true);
                        }
                    }

                    imagettftext($tmpImage, $fontSize, $this->_options['stwm_watermark_type']['text_angle'], $textBox['x'], $textBox['y'], $textColor, $this->font, $text);

                    imagecopyresampled($baseImage->getImage(), $tmpImage, 0, 0, 0, 0, $baseImage->getWidth(), $baseImage->getHeight(), $baseImage->getWidth(), $baseImage->getHeight());
                }
            }

            // Generate WaterImage
            if($this->_options['stwm_watermark_type']['enabled_image'])
            {
                $watermarkImagePath = $this->_options['stwm_watermark_type']['image_path'];

                $minSize = (!empty($this->_options['stwm_watermark_type']['min_size']) ? $this->_options['stwm_watermark_type']['min_size'] : 100);

                if($watermarkImagePath && (
                    $baseImage->getWidth() >= $minSize ||
                    $baseImage->getHeight() >= $minSize)
                )
                {
                    $watermarkImagePath = \XF::getRootDirectory() . (substr($watermarkImagePath, 0, 1) == '/' ? '' : '/') . $watermarkImagePath;

                    if(file_exists($watermarkImagePath) || is_readable($watermarkImagePath))
                    {
                        /** @var \XF\Image\Gd|\XF\Image\Imagick $watermarkImage */
                        $watermarkImage = $imageManager->imageFromFile($watermarkImagePath);

                        $watermarkRatio = $watermarkImage->getWidth() / $watermarkImage->getHeight();

                        /*
                        $newWidth = $watermarkImage->getWidth();
                        if($baseImage->getWidth() < $watermarkImage->getWidth() || $baseImage->getHeight() < $watermarkImage->getHeight())
                        {
                            if ($baseImage->getWidth() / $baseImage->getHeight() > $watermarkRatio)
                            {
                                $newWidth = $baseImage->getHeight() * $watermarkRatio;
                            }
                            else
                            {
                                $newWidth = $baseImage->getWidth();
                            }
                        }
                        */

                        $watermarkMaxWidth = $watermarkImage->getWidth() * 1.2;
                        $percentageRatio = $this->_options['stwm_watermark_type']['percentage'] / 100;
                        $percentageRatio = min(1, (!is_numeric($percentageRatio) ? 0.01 : max(0.01, $percentageRatio)));
                        $watermark_area = ($baseImage->getHeight() * $baseImage->getWidth()) * $percentageRatio;
                        $newWidth = round(sqrt($watermark_area / $watermarkRatio), 0);

                        // Fixes
                        if($newWidth > $watermarkMaxWidth)
                        {
                            $newWidth = $watermarkMaxWidth;
                        }
                        if($newWidth > $baseImage->getWidth())
                        {
                            $newWidth = $baseImage->getWidth();
                        }

                        $watermarkImage
                            ->setOpacity((100 - $this->_options['stwm_watermark_type']['image_opacity']) / 100)
                            ->resize($newWidth);

                        $offsetX = ($this->_options['stwm_watermark_type']['image_posX'] * ($baseImage->getWidth() - $watermarkImage->getWidth())) / 100;
                        $offsetY = ($this->_options['stwm_watermark_type']['image_posY'] * ($baseImage->getHeight() - $watermarkImage->getHeight())) / 100;

                        $baseImage->appendImageAt($offsetX, $offsetY, $watermarkImage->getImage());
                    }
                }
            }

            $qualitiy = $this->_options['stwm_availableImageTypes']['png_quality'];
            if($baseImage->getType() == IMAGETYPE_JPEG)
            {
                $qualitiy = $this->_options['stwm_availableImageTypes']['jpg_quality'];
            }

            @unlink($this->outputImage);
            $baseImage->save($this->outputImage, $baseImage->getType(), $qualitiy);

            if(file_exists($this->outputImage))
            {
                $this->status = true;
                $this->isNew = true;
            }

        }
    }

    private function getTextBox(&$fontSize, $angle, &$font, $text, $maxWidth, $maxHeight)
    {
        $bb = imagettfbbox($fontSize, $angle, $font, $text);

        $maxX = max($bb[0], $bb[2], $bb[4], $bb[6]);
        $minX = min($bb[0], $bb[2], $bb[4], $bb[6]);
        $x = $maxX - $minX;

        $maxY = max($bb[1], $bb[3], $bb[5], $bb[7]);
        $minY = min($bb[1], $bb[3], $bb[5], $bb[7]);
        $y = $maxY - $minY;

        $tX = ($this->_options['stwm_watermark_type']['text_posX'] * (($maxWidth < $x) ? 0 : $maxWidth - $x)) / 100;
        $tY = ($this->_options['stwm_watermark_type']['text_posY'] * (($maxHeight < $y) ? 0 : $maxHeight - $y)) / 100 ;

        // Reduziere Größe der Font
        if($x > $maxWidth || $y > $maxHeight)
        {
            $fontSize--;
            return $this->getTextBox($fontSize, $angle, $font, $text, $maxWidth, $maxHeight);
        }

        return [
            'left'   => abs($minX) - 1,
            'top'    => abs($minY) - 1,
            'width'  => $minX,
            'height' => $minY,
            'x' => $tX - $minX,
            'y' => $tY + $y - $maxY,
            'box'    => $bb
        ];
    }

    private function colorToRGB($hex)
    {
        $hex = strtolower($hex);

        //is rgbA? FUCK IT!
        if(strpos($hex, 'rgba') !== false)
        {
            preg_match('/^rgba?\((\d+),\s*(\d+),\s*(\d+)(?:,\s*(\d+(?:\.\d+)?))?\)$/i', $hex, $rgb);

            if($rgb)
            {
                return array($rgb[1], $rgb[2], $rgb[3]);
            }
        }

        if(strpos($hex, 'rgb') !== false)
        {
            preg_match('/^rgb\(\s*(\d+%?)\s*,\s*(\d+%?)\s*,\s*(\d+%?)\s*\)$/i', $hex, $rgb);

            if($rgb)
            {
                return array($rgb[1], $rgb[2], $rgb[3]);
            }
        }
        else
        {
            $hex = str_replace('#', '', $hex);

            if(utf8_strlen($hex) == 3)
            {
                $r = hexdec(substr($hex, 0, 1) . substr($hex, 0, 1));
                $g = hexdec(substr($hex, 1, 1) . substr($hex, 1, 1));
                $b = hexdec(substr($hex, 2, 1) . substr($hex, 2, 1));
            }
            else
            {
                $r = hexdec(substr($hex,0,2));
                $g = hexdec(substr($hex,2,2));
                $b = hexdec(substr($hex,4,2));
            }

            return array($r, $g, $b);
        }

        return array(0,0,0);
    }

    public function getWatermarkDataFilePath(\XF\Entity\Attachment $attachment)
    {
        return sprintf('%s/stwm_watermarks/%d/%d-%s.data',
            $this->_internalPath,
            floor($attachment['data_id'] / 1000),
            $attachment->Data['data_id'],
            $attachment->Data['file_hash']
        );
    }

    /**
     * @return \XF\Repository\Attachment
     */
    protected function getAttachmentRepo()
    {
        return \XF::app()->em()->getRepository('XF:Attachment');
    }

}